
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: DURABILITY, GameTooltip, LEVEL, PaperDollItemSlotButton_OnModifiedClick, PickupInventoryItem, ShoppingTooltip1, ShoppingTooltip2, ToggleCharacter, type, STATUS, TRANSMOG_NO_VALID_ITEMS_EQUIPPED

local floor = floor
local format = format
local GetDetailedItemLevelInfo = GetDetailedItemLevelInfo
local GetInventoryItemDurability = GetInventoryItemDurability
local GetInventoryItemLink = GetInventoryItemLink
local GetItemInfoInstant = GetItemInfoInstant
local next = next
local table = table

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(SixUI) -- AddOn, Border, Constants, Database
local M = A:NewModule("Display Durability")
local O = LibStub("LibDataBroker-1.1"):NewDataObject("SixUI Durability", {
	type = "data source",
	text = "0 %",
	value = "0",
	suffix = "%",
	label = "Durability",
})
local Social = LibStub("SixSocial-1.0")
local Tooltip = LibStub("SixTooltip-1.0"):CreateTooltip({
	cellMarginH = C.TOOLTIP_CELL_MARGIN_H,
	cellMarginV = C.TOOLTIP_CELL_MARGIN_V,
	columnJustification = {"LEFT", "LEFT", "RIGHT", "RIGHT"},
	font = C.TEXT_REGULAR,
	headerFont = C.TEXT_LARGE,
	key = "SixUI Durability",
	numColumns = 4,
	separatorHeight = C.TOOLTIP_SEPARATOR_HEIGHT,
})

local col = nil
local falsy = Social.falsy
local items = nil
local rev = nil
local upper = Social.upper
local worst = nil

M.Enum = {}

M.Enum.Item = {
	Index = "index",
	Level = "level",
	Link = "link",
	Name = "name",
	SlotId = "slotID",
	Status = "status",
	Value = "value",
}

M.Enum.Slot = {
	Head = INVSLOT_HEAD,
	Neck = INVSLOT_NECK,
	Shoulder = INVSLOT_SHOULDER,
	Chest = INVSLOT_CHEST,
	Waist = INVSLOT_WAIST,
	Legs = INVSLOT_LEGS,
	Feet = INVSLOT_FEET,
	Wrist = INVSLOT_WRIST,
	Hand = INVSLOT_HAND,
	Finger1 = INVSLOT_FINGER1,
	Finger2 = INVSLOT_FINGER2,
	Trinket1 = INVSLOT_TRINKET1,
	Trinket2 = INVSLOT_TRINKET2,
	Back = INVSLOT_BACK,
	MainHand = INVSLOT_MAINHAND,
	OffHand = INVSLOT_OFFHAND,
}

M.Enum.SlotText = {
	[M.Enum.Slot.Head] = INVTYPE_HEAD,
	[M.Enum.Slot.Neck] = INVTYPE_NECK,
	[M.Enum.Slot.Shoulder] = INVTYPE_SHOULDER,
	[M.Enum.Slot.Chest] = INVTYPE_CHEST,
	[M.Enum.Slot.Waist] = INVTYPE_WAIST,
	[M.Enum.Slot.Legs] = INVTYPE_LEGS,
	[M.Enum.Slot.Feet] = INVTYPE_FEET,
	[M.Enum.Slot.Wrist] = INVTYPE_WRIST,
	[M.Enum.Slot.Hand] = INVTYPE_HAND,
	[M.Enum.Slot.Finger1] = INVTYPE_FINGER .. " 1",
	[M.Enum.Slot.Finger2] = INVTYPE_FINGER .. " 2",
	[M.Enum.Slot.Trinket1] = INVTYPE_TRINKET .. " 1",
	[M.Enum.Slot.Trinket2] = INVTYPE_TRINKET .. " 2",
	[M.Enum.Slot.Back] = INVTYPE_CLOAK,
	[M.Enum.Slot.MainHand] = INVTYPE_WEAPONMAINHAND,
	[M.Enum.Slot.OffHand] = INVTYPE_WEAPONOFFHAND,
}

M.Enum.SlotFrame = {
	[M.Enum.Slot.Head] = CharacterHeadSlot,
	[M.Enum.Slot.Neck] = CharacterNeckSlot,
	[M.Enum.Slot.Shoulder] = CharacterShoulderSlot,
	[M.Enum.Slot.Chest] = CharacterChestSlot,
	[M.Enum.Slot.Waist] = CharacterWaistSlot,
	[M.Enum.Slot.Legs] = CharacterLegsSlot,
	[M.Enum.Slot.Feet] = CharacterFeetSlot,
	[M.Enum.Slot.Wrist] = CharacterWristSlot,
	[M.Enum.Slot.Hand] = CharacterHandsSlot,
	[M.Enum.Slot.Finger1] = CharacterFinger0Slot,
	[M.Enum.Slot.Finger2] = CharacterFinger1Slot,
	[M.Enum.Slot.Trinket1] = CharacterTrinket0Slot,
	[M.Enum.Slot.Trinket2] = CharacterTrinket1Slot,
	[M.Enum.Slot.Back] = CharacterBackSlot,
	[M.Enum.Slot.MainHand] = CharacterMainHandSlot,
	[M.Enum.Slot.OffHand] = CharacterSecondaryHandSlot,
}

M.Enum.Status = {
	Broken = 0,
	Breaking = 1,
	Normal = 2,
}

M.Enum.StatusColor = {
	[M.Enum.Status.Normal] = CreateColor(0.07, 0.93, 0.07),
	[M.Enum.Status.Breaking] = CreateColor(1, 0.82, .18),
	[M.Enum.Status.Broken] = CreateColor(0.93, 0.07, 0.07),
}

local ITEM_DEFAULTS = setmetatable({
	[M.Enum.Item.Index] = 0,
	[M.Enum.Item.Level] = 0,
	[M.Enum.Item.Link] = nil,
	[M.Enum.Item.Name] = UNKNOWN,
	[M.Enum.Item.SlotId] = 0,
	[M.Enum.Item.Status] = M.Enum.Status.Normal,
	[M.Enum.Item.Value] = 100,
}, {
	__concat = function (a, b)
		return Social.merge(a, b)
	end,
	__lt = function (a, b)
		if (a.value ~= b.value) then
			return (a.value < b.value)
		else
			return (a.index < b.index)
		end
	end,
})

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

function M:OnEnable()
	self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED", "Update")
	self:RegisterEvent("UPDATE_INVENTORY_ALERTS", "Update")
	self:RegisterEvent("UPDATE_INVENTORY_DURABILITY", "Update")
	self:RegisterMessage("SIXUI_CLUTTER_STATE_CHANGED", "Update")
	self:Update()
end

function M:Update(event, ...)
	if event == "GET_ITEM_INFO_RECEIVED" then
		self:UnregisterEvent("GET_ITEM_INFO_RECEIVED")
	end

	if event == "SIXUI_CLUTTER_STATE_CHANGED" then
		self.isClutterKeyDown = ...
	else
		items, worst = self:GetItems()
		O.text = format("%s%%", worst.value)
		O.value = worst.value
	end

	Tooltip:Refresh()
end

function M:GetItems()
	local items, worst = {}, ITEM_DEFAULTS

	for _, slotID in next, M.Enum.Slot do
		local item = self:GetItem(slotID, slotID)

		if item then
			table.insert(items, item)

			if falsy(item.name) then
				self:RegisterEvent("GET_ITEM_INFO_RECEIVED", "Update")
			end

			if item < worst then
				worst = item
			end
		end
	end

	return items, worst
end

function M:GetItem(index, slotID)
	local itemLink = GetInventoryItemLink("player", slotID)
	if (not itemLink) then return end
	local current, maximum = GetInventoryItemDurability(slotID)
	local value = maximum and floor(current / maximum * 100) or 100
	local status = value > 10 and M.Enum.Status.Normal or value > 0 and M.Enum.Status.Breaking or M.Enum.Status.Broken

	return ITEM_DEFAULTS .. {
		[M.Enum.Item.Index] = index,
		[M.Enum.Item.Level] = GetDetailedItemLevelInfo(itemLink) or 0,
		[M.Enum.Item.Link] = itemLink,
		[M.Enum.Item.Name] = GetItemInfoInstant(itemLink),
		[M.Enum.Item.SlotId] = slotID,
		[M.Enum.Item.Status] = status,
		[M.Enum.Item.Value] = value,
	}
end

-------------------------------------------------------------------------------
-- DataObject
-------------------------------------------------------------------------------

function O.OnEnter(frame)
	if frame then
		Tooltip:UpdateConfig({anchor = frame})
	end

	Tooltip:Show()
end

function O.OnClick(frame, button)
	ToggleCharacter("PaperDollFrame")
end

-------------------------------------------------------------------------------
-- Tooltip
-------------------------------------------------------------------------------

function Tooltip:OnShow()
	if not items[1] then
		self:AddHeader(DURABILITY)
		self:AddSeparator(C.TOOLTIP_SEPARATOR_HEIGHT)
		self:AddLine(TRANSMOG_NO_VALID_ITEMS_EQUIPPED)
	else
		self:AddItemsHeader()
		self:AddItems()
	end
end

local function Header_OnClick(self, key, button)
	A:HandleClick({
		{mouseButton = C.Enum.MouseButton.Left,   callback = function () col, rev = key, nil end},
		{mouseButton = C.Enum.MouseButton.Middle, callback = function () col, rev = nil, nil end},
		{mouseButton = C.Enum.MouseButton.Right,  callback = function () col, rev = key, key end},
	}, button)

	Tooltip:Refresh()
end

function Tooltip:AddItemsHeader()
	if M.isClutterKeyDown then
		local line = self:AddHeader("Slot", "Item", LEVEL, STATUS)

		self:SetCellScript(line, 1, "OnMouseUp", Header_OnClick, "index")
		self:SetCellScript(line, 2, "OnMouseUp", Header_OnClick, "name")
		self:SetCellScript(line, 3, "OnMouseUp", Header_OnClick, "level")
		self:SetCellScript(line, 4, "OnMouseUp", Header_OnClick, "value")
	else
		local line = self:AddHeader()

		self:SetCell(line, 1, DURABILITY, nil, "LEFT", 4)
	end

	self:AddSeparator(C.TOOLTIP_SEPARATOR_HEIGHT)
end

local function SortItems(a, b)
	if rev then a, b = b, a end
	local x, y = a[col], b[col]

	if not x or not y then
		return not not x
	elseif x ~= y and type(x) == "number" and type(y) == "number" then
		return x < y
	elseif x ~= y and type(x) == "string" and type(y) == "string" then
		return upper(x or "ZZZ") < upper(y or "ZZZ")
	else
		return a < b
	end
end

function Tooltip:AddItems()
	if col then
		table.sort(items, SortItems)
	else
		table.sort(items)
	end

	for _, item in next, items do
		self:AddItem(item)
	end
end

local function Item_OnEnter(self, item)
	GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
	GameTooltip:SetInventoryItem("player", item.slotID)
	ShoppingTooltip1:Hide()
	ShoppingTooltip2:Hide()
end

local function Item_OnLeave(self)
	if GameTooltip:IsOwned(self) then
		GameTooltip:Hide()
	end
end

local function Item_OnClick(self, item, button)
	A:HandleClick({
		{mouseButton = C.Enum.MouseButton.Left,  modifierKey = C.Enum.ModifierKey.None,  callback = function () PickupInventoryItem(item.slotID) end},
		{mouseButton = C.Enum.MouseButton.Left,  modifierKey = C.Enum.ModifierKey.Shift, callback = function () A.OpenEditBox(); PaperDollItemSlotButton_OnModifiedClick(M.Enum.SlotFrame[item.slotID], button) end},
		{mouseButton = C.Enum.MouseButton.Right, modifierKey = C.Enum.ModifierKey.Shift, callback = function () PaperDollItemSlotButton_OnModifiedClick(M.Enum.SlotFrame[item.slotID], button) end},
	}, button)
end

function Tooltip:AddItem(item)
	local line = self:AddLine(
		M.Enum.SlotText[item.slotID],
		item.link,
		item.level,
		M.Enum.StatusColor[item.status]:WrapTextInColorCode(format("%d%%", item.value))
	)

	self:SetLineScript(line, "OnEnter", Item_OnEnter, item)
	self:SetLineScript(line, "OnLeave", Item_OnLeave)
	self:SetLineScript(line, "OnMouseDown", Item_OnClick, item)
end
